# Copyright (c) 2014, GPy authors (see AUTHORS.txt).
# Licensed under the BSD 3-clause license (see LICENSE.txt)

# from . import base_plots
# from . import models_plots
# from . import priors_plots
# from . import variational_plots
# from . import kernel_plots
# from . import dim_reduction_plots
# from . import mapping_plots
# from GPy.plotting.gpy_plot import Tango
# from . import visualize
# from . import latent_space_visualizations
# from . import inference_plots
# from . import maps
# from . import img_plots
# from .ssgplvm import SSGPLVM_plot


from .util import align_subplot_array, align_subplots, fewerXticks, removeRightTicks, removeUpperTicks
from . import controllers, base_plots
