#include <math.h>
double DiracDelta(double x);
double DiracDelta(double x, int foo);

double sinc(double x);
double sinc_grad(double x);

double erfcx(double x);
double ln_diff_erf(double x0, double x1);
