# Copyright (c) 2012, GPy authors (see AUTHORS.txt).
# Licensed under the BSD 3-clause license (see LICENSE.txt)
"""
Introduction
^^^^^^^^^^^^

A variety of utility functions including matrix operations and quick access to test datasets.

"""

from . import linalg
from . import misc
from . import squashers
from . import warping_functions
from . import mocap
from . import decorators
from . import classification
from . import subarray_and_sorting
from . import diag
from . import initialization
from . import multioutput
from . import parallel
from . import functions
from . import cluster_with_offset
from . import input_warping_functions
